{
	TCalculator
	-----------

		Copyright Pablo Pissanetzky
		http://www.pablop.demon.co.uk
		pablop@bigfoot.com
}

unit Calculator;

interface

uses
  Classes , SysUtils;

type
  TCharClass = ( ccPeriod , ccNumber , ccQuote , ccAlpha , ccOParen ,
    ccCParen , ccMinus , ccOp , ccCustom , ccComma , ccSpace , ccEqual , ccLess,
    ccGreater , ccOther  );

  TValueType = ( vtNumber , vtString , vtIdentifier , vtEmpty );
  TValueTypes = set of TValueType;

  ECalculator = class( Exception ) end;

  TCalculator = class( TObject )
  private
    FCustomOperators : string;
    function CharClass( P : PChar ) : TCharClass;
    function Tokenize( const Text : string ) : TStrings;
    function Syntax( Tokens : TStrings ) : TStrings;
  protected
    function ResolveIdentifier( const Identifier : string;
      var ValueType : TValueType; var Value : string ) : Boolean; virtual;
    function LookupFunction( const FunctionName : string;
      var MinArgCount , MaxArgCount : Integer; var ValueTypes : TValueTypes ) : Boolean; virtual;
    function SolveFunction( const FunctionName : string;
      Arguments : TStrings; var ReturnType : TValueType; var Return : string ) : Boolean; virtual;
    procedure ResolveArguments( Arguments : TStrings; ValueTypes : TValueTypes );
    function ArgumentType( Index : Integer; Arguments : TStrings ) : TValueType;
    procedure ReplaceEmptysWith( Arguments : TStrings; const S : string );
  public
    property CustomOperators : string read FCustomOperators write FCustomOperators;
    function Solve( const Formula : string; var ReturnType : TValueType  ) : string;
  end;

  function EqualStr( const A , B : string ) : Boolean;
  function StringEval( const S : string; Targets : array of PChar ) : Integer;

implementation

uses
  Windows , DList;

//------------------------------------------------------------------------------

type

  TSymbolInfo = record
    Name        : string;
    Nest        : Integer;
    SymbolType  : Integer;
  end;

  PSymbolNode = ^ TSymbolNode;
  TSymbolNode = record
    ListNode    : TDListNode;
    SymbolInfo  : TSymbolInfo;
  end;

  TSymbolStack = class( TObject )
  private
    FDList  : TDList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure PushSymbol( const Name : string; Nest , SymbolType : Integer );
    function PopSymbol( Nest : Integer; var SymbolInfo : TSymbolInfo ) : Boolean;
    function Count : Integer;
  end;

  constructor TSymbolStack.Create;
  begin
    inherited;
    FDList := TDlist.Create( SizeOf( TSymbolInfo ) , nil , True );
  end;

  destructor TSymbolStack.Destroy;
  begin
    FDList.Free;
    inherited;
  end;

  procedure TSymbolStack.PushSymbol( const Name : string; Nest , SymbolType : Integer );
  var
    NewNode : PSymbolNode;
  begin
    New( NewNode );
    with NewNode^ do
      begin
        SymbolInfo.Name := Name;
        SymbolInfo.Nest := Nest;
        SymbolInfo.SymbolType := SymbolType;
      end;
    FDList.InsertNodeFirst( PDListNode( NewNode ) );
  end;

  function TSymbolStack.PopSymbol( Nest : Integer; var SymbolInfo : TSymbolInfo ) : Boolean;
  var
    Node : PSymbolNode;
  begin
    Result := False;
    Node := PSymbolNode( FDList.First );
    if Node <> nil then
      begin
        Result := Node.SymbolInfo.Nest = Nest;
        if Result then
          begin
            SymbolInfo := Node.SymbolInfo;
            Dispose( FDList.RemoveFirst );
          end;
      end;
  end;

  function TSymbolStack.Count : Integer;
  begin
    Result := FDList.Count;
  end;

//------------------------------------------------------------------------------

const
  TokenizerFSA : array [ 0..17 , TCharClass ] of Integer =
  (
                  {   .       0-9     "     A-Z     (     )     -     +*/     custom  ,     SP  , =   , <   , >   , other }
{00} {start}      (   1   ,   2    ,  3   , 4     , 5   , 6   , 7   , 8     , 11    ,  9  , 0   , 12  , 13  , 14  , 500   ),
{01} {float}      (   500 ,   1    ,  100 , 100   , 100 , 100 , 100 , 100   , 100   , 100 , 100 , 100 , 100 , 100 , 500   ),
{02} {int}        (   1   ,   2    ,  100 , 100   , 100 , 100 , 100 , 100   , 100   , 100 , 100 , 100 , 100 , 100 , 500   ),
{03} {literal}    (   3   ,   3    ,  10  , 3     , 3   , 3   , 3   , 3     , 3     , 3   , 3   , 3   , 3   , 3   , 3     ),
{04} {id}         (   102 ,   4    ,  102 , 4     , 102 , 102 , 102 , 102   , 102   , 102 , 102 , 102 , 102 , 102 , 500   ),
{05} {oparen}     (   103 ,   103  ,  103 , 103   , 103 , 103 , 103 , 103   , 103   , 103 , 103 , 103 , 103 , 103 , 500   ),
{06} {cparen}     (   104 ,   104  ,  104 , 104   , 104 , 104 , 104 , 104   , 104   , 104 , 104 , 104 , 104 , 104 , 500   ),
{07} {minus}      (   105 ,   105  ,  105 , 105   , 105 , 105 , 105 , 105   , 105   , 105 , 105 , 105 , 105 , 105 ,500    ),
{08} {op}         (   106 ,   106  ,  106 , 106   , 106 , 106 , 106 , 106   , 106   , 106 , 106 , 106 , 106 , 106 , 500   ),
{09} {comma}      (   107 ,   107  ,  107 , 107   , 107 , 107 , 107 , 107   , 107   , 107 , 107 , 107 , 107 , 107 , 500   ),
{10} {literal}    (   101 ,   101  ,  101 , 101   , 101 , 101 , 101 , 101   , 101   , 101 , 101 , 101 , 101 , 101 , 500   ),
{11} {custom}     (   108 ,   108  ,  108 , 108   , 108 , 108 , 108 , 108   , 108   , 108 , 108 , 108 , 108 , 108 , 500   ),
{12} {equal}      (   109 ,   109  ,  109 , 109   , 109 , 109 , 109 , 109   , 109   , 109 , 109 , 109 , 109 , 109 , 500   ),
{13} {less}       (   110 ,   110  ,  110 , 110   , 110 , 110 , 110 , 110   , 110   , 110 , 110 , 15  , 110 , 17  , 500   ),
{14} {greater}    (   111 ,   111  ,  111 , 111   , 111 , 111 , 111 , 111   , 111   , 111 , 111 , 16  , 111 , 111 , 500   ),
{15} {<=}         (   112 ,   112  ,  112 , 112   , 112 , 112 , 112 , 112   , 112   , 112 , 112 , 112 , 112 , 112 , 500   ),
{16} {>=}         (   113 ,   113  ,  113 , 113   , 113 , 113 , 113 , 113   , 113   , 113 , 113 , 113 , 113 , 113 , 500   ),
{17} {<>}         (   114 ,   114  ,  114 , 114   , 114 , 114 , 114 , 114   , 114   , 114 , 114 , 114 , 114 , 114 , 500   )
  );

function EqualStr( const A , B : string ) : Boolean;
begin
  Result := CompareText( A , B ) = 0;
end;

function StringEval( const S : string; Targets : array of PChar ) : Integer;
var
  I : Integer;
begin
  Result := -1;
  for I := Low( Targets ) to High( Targets ) do
    if StrIComp( PChar( S ) , Targets[ I ] ) = 0 then
      begin
        Result := I - Low( Targets );
        Break;
      end;
end;

function TCalculator.CharClass( P : PChar ) : TCharClass;
begin
  case P^ of
    '.'         : Result := ccPeriod;
    '0'..'9'    : Result := ccNumber;
    '"'         : Result := ccQuote;
    '('         : Result := ccOParen;
    ')'         : Result := ccCParen;
    ','         : Result := ccComma;
    'A'..'Z',
    'a'..'z'    : Result := ccAlpha;
    '-'         : Result := ccMinus;
    '+','*','/' : Result := ccOp;
    '='         : Result := ccEqual;
    '<'         : Result := ccLess;
    '>'         : Result := ccGreater;
    ' ',#9,#10,
    #13,#0      : Result := ccSpace;
    else
      if Pos( Char( P^ ) , FCustomOperators ) <> 0 then
        Result := ccCustom
      else
        Result := ccOther;
  end;
end;

function TCalculator.Tokenize( const Text : string ) : TStrings;
const
  ErrorState = 500;
var
  State , NextState : Integer;
  Token : string;
  P , PS : PChar;
begin
  State := 0;
  P := PChar( Text );
  PS := P;
  Result := TStringList.Create;
  try
    while True do
      begin
        NextState := TokenizerFSA[ State , CharClass( P ) ];
        if NextState = ErrorState then
          raise ECalculator.Create( 'Invalid character "' + P^ + '"' )
        else if NextState in [ 100..114 ] then
          begin
            SetString( Token , PS , P - PS );
            Result.AddObject( Trim( Token ) , Pointer( NextState ) );
            State := 0;
            PS := P;
            if P^ = #0 then
              Break;
          end
        else
          begin
            if P^ = #0 then
              begin
                if State = 3 then { unterminated string literal }
                  begin
                    SetString( Token , PS , P - PS );
                    Result.AddObject( Trim( Token + '"' ) , Pointer( 101 ) );
                  end;
                Break;
              end;
            State := NextState;
            P := CharNext( P );
          end;
      end;
  except
    Result.Free;
    raise;
  end;
end;

function TCalculator.Syntax( Tokens : TStrings ) : TStrings;
var
  SymbolStack : TSymbolStack;
  Nest  : Integer;

  procedure PopSymbol;
  var
    SymbolInfo : TSymbolInfo;
  begin
    while SymbolStack.Count > 0 do
      begin
        if SymbolStack.PopSymbol( Nest , SymbolInfo ) then
          Result.AddObject( SymbolInfo.Name , Pointer( SymbolInfo.SymbolType ) )
        else
          Break;
      end;
  end;

  function TokenType( Index : Integer ) : Integer;
  begin
    if ( Index < Tokens.Count ) and ( Index >= 0 ) then
      Result := Longint( Tokens.Objects[ Index ] )
    else
      Result := 500;
  end;

  procedure PushSymbol( const Symbol : string; SymbolType : Integer );
  begin
    SymbolStack.PushSymbol( Symbol , Nest , SymbolType );
  end;

  procedure PutToken( const Token : string; TType : Integer );
  begin
    Result.AddObject( Token , Pointer( TType ) );
  end;

var
  I : Integer;

  { Symbol types : 200 - operator , 201 - function call , 202 - arg start }

begin
  SymbolStack := TSymbolStack.Create;
  try
    Result := TStringList.Create;
    try
      Nest := 0;
      for I := 0 to Pred( Tokens.Count ) do
        case TokenType( I ) of

          100 { number }, 101 { literal }:
            begin
              PutToken( Tokens[ I ] , TokenType( I ) );
              PopSymbol;
            end;

          102 { id } :
            begin
              if TokenType( I + 1 ) = 103 then
                begin
                  PushSymbol( Tokens[ I ] , 201 );
                  PutToken( 'ARGS' , 202 );
                end
              else
                begin
                  PutToken( Tokens[ I ] , TokenType( I ) );
                  PopSymbol;
                end;
            end;

          103 : { ( }
            Inc( Nest );

          104 : { ) }
            begin
              Dec( Nest );
              if Nest < 0 then
                raise ECalculator.Create( 'Too many closing parenthesis' );
              PopSymbol;
            end;

          105 : { - }
            begin
              case TokenType( I - 1 ) of
                500,103,105..114: PushSymbol( '--' , 200 )
                else PushSymbol( '-' , 200 );
              end;
            end;

          106 : { +*/ }
            PushSymbol ( Tokens[ I ] , 200 );

          107 : { , }
            begin
            end;

          108 : { custom operator }
            PushSymbol ( Tokens[ I ] , 200 );

          109..114 : { = < > <= >= <>}
            PushSymbol( Tokens[ I ] , 200 );

        end;
      if Nest <> 0 then
        raise ECalculator.Create( 'Missing closing parenthesis' );
      if SymbolStack.Count > 0 then
        raise ECalculator.Create( 'Invalid expression' );
    except
      Result.Free;
      Result := nil;
      raise;
    end;
  finally
    SymbolStack.Free;
  end;
end;

function TCalculator.Solve( const Formula : string; var ReturnType : TValueType ) : string;
var
  I , K , FirstArg , MinArgCount , MaxArgCount : Integer;
  P : PChar;
  Value : string;
  ValueType : TValueType;
  ValueTypes : TValueTypes;
  Tokens , FSyntax , Parameters : TStrings;
  TType : Integer;
begin
  Parameters := TStringList.Create;
  try
    Tokens := Tokenize( Formula );
    try
      FSyntax := Syntax( Tokens );
      try
        I := 0;
        while I < FSyntax.Count do
          begin
            TType := Longint( FSyntax.Objects[ I ] );
            case TType of

              100 : { number }
                begin
                  { ignore it }
                  FSyntax.Objects[ I ] := Pointer( vtNumber );
                end;

              101 : { string literal }
                begin
                  P := PChar( FSyntax[ I ] );
                  FSyntax[ I ] := AnsiExtractQuotedStr( P , '"' );
                  FSyntax.Objects[ I ] := Pointer( vtString);
                end;

              102 : { identifier }
                begin
                  FSyntax.Objects[ I ] := Pointer( vtIdentifier );
                end;

              202 : { argument start marker }
                begin
                  { ignored }
                end;

              200 , 201 : { operator or function name }
                begin
                  FirstArg := 0; // To get rid of warning
                  // First, find out if the function exists, how many arguments it
                  // expects and the possible types of the arguments
                  if not LookupFunction( FSyntax[ I ] , MinArgCount, MaxArgCount , ValueTypes ) then
                    raise ECalculator.Create( 'Invalid function/operator "' + FSyntax[ I ] + '"' );
                  if ( MinArgCount < 0 ) or ( MaxArgCount < 0 ) then
                    raise ECalculator.Create( 'Invalid function/operator argument count "'+ FSyntax[ I ] + '"' );
                  if MaxArgCount < MinArgCount then
                    raise ECalculator.Create( 'Invalid function/operator argument count "'+ FSyntax[ I ] + '"' );

                  if TType = 200 then
                    begin
                      if MinArgCount <> MaxArgCount then
                        raise ECalculator.Create( 'Operator may not have variable arguments' );
                      FirstArg := I - MinArgCount;
                      if  FirstArg < 0 then
                        raise ECalculator.Create( 'Invalid number of arguments for function/operator "' + FSyntax[ I ] + '"' );
                    end
                  else if TType = 201 then // if it is a function call, we should have an argument marker
                    begin
                      FirstArg := I - 1;
                      while ( FirstArg >=0 ) and ( Longint( FSyntax.Objects[ FirstArg ] ) <> 202 ) do
                        Dec( FirstArg );
                      // ARGS marker was not found
                      if FirstArg < 0 then
                        raise ECalculator.Create( 'Invalid number of arguments for function "' + FSyntax[ I ] + '"' );
                      // There is no ARGS marker where there should be one
                      if ( I - FirstArg - 1 ) < MinArgCount then
                        raise ECalculator.Create( 'Not enough arguments in call to function "' + FSyntax[ I ] + '"' );
                      if ( I - FirstArg - 1 ) > MaxArgCount then
                        raise ECalculator.Create( 'Too many arguments in call to function "' + FSyntax[ I ] + '"' );

                      // Get rid of the args marker
                      FSyntax.Delete( FirstArg );
                      // Adjust the indices for the deletion
                      Dec( I );
                    end;

                  Parameters.Clear;
                  // Verify that all the arguments are of the required type,
                  // and populate the Parameters string list with them
                  for K := FirstArg to Pred( I ) do
                    if not ( TValueType( FSyntax.Objects[ K ] ) in ValueTypes ) then
                      raise ECalculator.Create( 'Invalid argument type for function/operator "' + FSyntax[ I ] + '"' )
                    else
                      Parameters.AddObject( FSyntax[ K ] , FSyntax.Objects[ K ] );

                  // Solve the function with the given arguments
                  if not SolveFunction( FSyntax[ I ] , Parameters , ValueType , Value ) then
                    raise ECalculator.Create( 'Unable to solve function/operator "' + FSyntax[ I ] + '"' );
                  // Put the return value on the stack and get rid of all
                  // the other stuff
                  FSyntax[ FirstArg ] := Value;
                  FSyntax.Objects[ FirstArg ] := Pointer( ValueType );
                  K := FirstArg + 1;
                  while K <= I do
                    begin
                      FSyntax.Delete( K );
                      Dec( I );
                    end;
                  Dec( I );
                end;
            end;
            Inc( I );
          end;
        if FSyntax.Count <> 1 then
          raise ECalculator.Create( 'Invalid formula syntax' );
        if TValueType( FSyntax.Objects[ 0 ] ) = vtIdentifier then
          begin
            if not ResolveIdentifier( FSyntax[ 0 ] , ReturnType , Result ) then
              raise ECalculator.Create( 'Invalid identifier "' + FSyntax[ 0 ] + '"' );
            if ReturnType = vtIdentifier then
              raise ECalculator.Create( 'Unresolved identifier "' + FSyntax[ 0 ] + '"' );
          end
        else
          begin
            ReturnType := TValueType( FSyntax.Objects[ 0 ] );
            Result := FSyntax[ 0 ];
          end;
      finally
        FSyntax.Free;
      end;
    finally
      Tokens.Free;
    end;
  finally
    Parameters.Free;
  end;
end;

function TCalculator.ResolveIdentifier( const Identifier : string;
  var ValueType : TValueType; var Value : string ) : Boolean;
begin
  Result := False;
end;

function TCalculator.LookupFunction( const FunctionName : string;
  var MinArgCount , MaxArgCount : Integer; var ValueTypes : TValueTypes ) : Boolean;
begin
  Result := False;
  case StringEval( FunctionName ,
  [ '--' , '+' , '*' , '/' , '-' , 'TRUNC' , 'ROUND' , 'ABS' , '=' , '<' , '>' , '<=' , '>=' ,'<>' ] ) of
  0:  {--}
    begin
      MinArgCount := 1;
      MaxArgCount := 1;
      ValueTypes := [ vtIdentifier , vtNumber , vtEmpty ];
      Result := True;
    end;
  1:  {+}
    begin
      MinArgCount := 2;
      MaxArgCount := 2;
      ValueTypes := [ vtIdentifier , vtNumber , vtString , vtEmpty ];
      Result := True;
    end;
  2 , 3 , 4 :  {* / - }
    begin
      MinArgCount := 2;
      MaxArgCount := 2;
      ValueTypes := [ vtIdentifier , vtNumber , vtEmpty ];
      Result := True;
    end;
  5:  { TRUNC }
    begin
      MinArgCount := 1;
      MaxArgCount := 1;
      ValueTypes := [ vtIdentifier , vtNumber , vtEmpty ];
      Result := True;
    end;
  6:  { ROUND }
    begin
      MinArgCount := 1;
      MaxArgCount := 1;
      ValueTypes := [ vtIdentifier , vtNumber , vtEmpty ];
      Result := True;
    end;
  7: {ABS}
    begin
      MinArgCount := 1;
      MaxArgCount := 1;
      ValueTypes := [ vtIdentifier , vtNumber , vtEmpty ];
      Result := True;
    end;
  8 , 9 , 10 , 11 , 12 , 13 : {= < > <= >= <> }
    begin
      MinArgCount := 2;
      MaxArgCount := 2;
      ValueTypes := [ vtIdentifier , vtNumber , vtString , vtEmpty ];
      Result := True;
    end;
  end;
end;

function TCalculator.ArgumentType( Index : Integer; Arguments : TStrings ) : TValueType;
begin
  Result := TValueType( Arguments.Objects[ Index ] );
end;

procedure TCalculator.ResolveArguments( Arguments : TStrings; ValueTypes : TValueTypes );
var
  I : Integer;
  ValueType : TValueType;
  Value     : string;
begin
  for I := 0 to Pred( Arguments.Count ) do
    begin
      if ArgumentType( I , Arguments ) = vtIdentifier then
        begin
          if not ResolveIdentifier( Arguments[ I ] , ValueType , Value ) then
            raise ECalculator.Create( 'Invalid identifier "' + Arguments[ I ] + '"' );
          if ValueType = vtIdentifier then
            raise ECalculator.Create( 'Unresolved identifier "' + Arguments[ I ] + '"' );
          if not ( ValueType in ValueTypes ) then
            raise ECalculator.Create( 'Invalid argument type "' + Arguments[ I ] + '"' );
          Arguments[ I ] := Value;
          Arguments.Objects[ I ] := Pointer( ValueType );
        end;
    end;
end;

procedure TCalculator.ReplaceEmptysWith( Arguments : TStrings; const S : string );
var
  I : Integer;
begin
  for I := 0 to Pred( Arguments.Count ) do
    if ArgumentType( I ,Arguments ) = vtEmpty then
      Arguments[ I ] := S;
end;

function TCalculator.SolveFunction( const FunctionName : string;
  Arguments : TStrings; var ReturnType : TValueType; var Return : string ) : Boolean;

  function EmptyCount : Integer;
  var
    I : Integer;
  begin
    Result := 0;
    for I := 0 to Pred( Arguments.Count ) do
      if ArgumentType( I , Arguments ) = vtEmpty then
        Inc( Result );
  end;

  function AllEmpty : Boolean;
  begin
    Result := EmptyCount = Arguments.Count;
  end;


var
  BoolResult  : Boolean;
  EvaluateAs  : TValueType;
  FunctionNo   : Integer;
begin
  Result := False;
  FunctionNo := StringEval( FunctionName ,
    [ '--','*','/','-','+','TRUNC','ROUND','ABS','=','<','>','<=','>=','<>'] );
  case  FunctionNo of
  0:  {--}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtNumber ] );
      ReplaceEmptysWith( Arguments , '0' );
      Result := True;
      Return := FloatToStr( - StrToFloat( Arguments[ 0 ] ) );
      ReturnType := vtNumber;
    end;
  1:  {*}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtNumber ] );
      ReplaceEmptysWith( Arguments , '0' );
      Return := FloatToStr( StrToFloat( Arguments[ 0 ] ) * StrToFloat( Arguments[ 1 ] ) );
      ReturnType := vtNumber;
      Result := True;
    end;
  2:  {/}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtNumber ] );
      ReplaceEmptysWith( Arguments , '0' );
      Return := FloatToStr( StrToFloat( Arguments[ 0 ] ) / StrToFloat( Arguments[ 1 ] ) );
      ReturnType := vtNumber;
      Result := True;
    end;
  3:  {-}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtNumber ] );
      ReplaceEmptysWith( Arguments , '0' );
      Return := FloatToStr( StrToFloat( Arguments[ 0 ] ) - StrToFloat( Arguments[ 1 ] ) );
      ReturnType := vtNumber;
      Result := True;
    end;
  4:  {+}
    begin
      ResolveArguments( Arguments , [ vtString , vtEmpty , vtNumber ] );
      if AllEmpty then
        begin
          Result := True;
          Return := '0';
          ReturnType := vtNumber;
        end
      else
        begin
          if EmptyCount = 1 then
            begin
              Result := True;
              if ( ArgumentType( 0 , Arguments ) = vtString ) or  ( ArgumentType( 1 , Arguments ) = vtString ) then
                begin
                  ReplaceEmptysWith( Arguments , '' );
                  ReturnType := vtString;
                end
              else
                begin
                  ReplaceEmptysWith( Arguments , '0' );
                  ReturnType := vtNumber;
                end;
            end
          else
            begin
              Result := ArgumentType( 0 , Arguments ) = ArgumentType( 1 , Arguments );
              ReturnType := ArgumentType( 0 , Arguments );
            end;

          if Result then
            begin
              case ReturnType of
                vtString : Return := Arguments[ 0 ] + Arguments[ 1 ];
                vtNumber : Return := FloatToStr( StrToFloat( Arguments[ 0 ] ) + StrToFloat( Arguments[ 1 ] ) );
              end;
            end;
        end;
    end;
  5:  {TRUNC}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtNumber ] );
      ReplaceEmptysWith( Arguments , '0' );
      Result := True;
      ReturnType := vtNumber;
      Return := IntToStr( Trunc( StrToFloat( Arguments[ 0 ] ) ) );
    end;
  6:  {ROUND}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtNumber ] );
      ReplaceEmptysWith( Arguments , '0' );
      Result := True;
      ReturnType := vtNumber;
      Return := IntToStr( Round( StrToFloat( Arguments[ 0 ] ) ) );
    end;
  7:  {ABS}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtNumber ] );
      ReplaceEmptysWith( Arguments , '0' );
      Result := True;
      ReturnType := vtNumber;
      Return := FloatToStr( Abs( StrToFloat( Arguments[ 0 ] ) ) );
    end;
  8,9,10,11,12,13:  {= < > <= >= <>}
    begin
      ResolveArguments( Arguments , [ vtEmpty , vtString , vtNumber ] );
      ReturnType := vtString;
      Result := True;
      if EmptyCount = 1 then
        begin
          Result := True;
          if ( ArgumentType( 0 , Arguments ) = vtString ) or  ( ArgumentType( 1 , Arguments ) = vtString ) then
            begin
              ReplaceEmptysWith( Arguments , '' );
              EvaluateAs := vtString;
            end
          else
            begin
              ReplaceEmptysWith( Arguments , '0' );
              EvaluateAs := vtNumber;
            end;
        end
      else
        begin
          if ArgumentType( 0 , Arguments ) = ArgumentType( 1 , Arguments ) then
            EvaluateAs := ArgumentType( 0 , Arguments )
          else
            EvaluateAs := vtString;
        end;
      BoolResult := True; // Just to suppress warning
      case FunctionNo of
        8 : {=}
          case EvaluateAs of
            vtEmpty   : BoolResult := True;
            vtString  : BoolResult := Arguments[ 0 ] = Arguments[ 1 ];
            vtNumber  : BoolResult := StrToFloat( Arguments[ 0 ] ) = StrToFloat( Arguments[ 1 ] );
          end;
        9 : {<}
          case EvaluateAs of
            vtEmpty   : BoolResult := False;
            vtString  : BoolResult := Arguments[ 0 ] < Arguments[ 1 ];
            vtNumber  : BoolResult := StrToFloat( Arguments[ 0 ] ) < StrToFloat( Arguments[ 1 ] );
          end;
        10: {>}
          case EvaluateAs of
            vtEmpty   : BoolResult := False;
            vtString  : BoolResult := Arguments[ 0 ] > Arguments[ 1 ];
            vtNumber  : BoolResult := StrToFloat( Arguments[ 0 ] ) > StrToFloat( Arguments[ 1 ] );
          end;
        11: {<=}
          case EvaluateAs of
            vtEmpty   : BoolResult := True;
            vtString  : BoolResult := Arguments[ 0 ] <= Arguments[ 1 ];
            vtNumber  : BoolResult := StrToFloat( Arguments[ 0 ] ) <= StrToFloat( Arguments[ 1 ] );
          end;
        12: {>=}
          case EvaluateAs of
            vtEmpty   : BoolResult := True;
            vtString  : BoolResult := Arguments[ 0 ] >= Arguments[ 1 ];
            vtNumber  : BoolResult := StrToFloat( Arguments[ 0 ] ) >= StrToFloat( Arguments[ 1 ] );
          end;
        13: {<>}
          case EvaluateAs of
            vtEmpty   : BoolResult := False;
            vtString  : BoolResult := Arguments[ 0 ] <> Arguments[ 1 ];
            vtNumber  : BoolResult := StrToFloat( Arguments[ 0 ] ) <> StrToFloat( Arguments[ 1 ] );
          end;
      end;

      if BoolResult then Return := 'True'
      else Return := 'False';
    end;
  end;
end;


end.

