unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls , Calculator;

type

  TTestCalculator = class( TCalculator )
  protected
    function ResolveIdentifier( const Identifier : string;
      var ValueType : TValueType; var Value : string ) : Boolean; override;
    function LookupFunction( const FunctionName : string;
      var MinArgCount , MaxArgCount : Integer; var ValueTypes : TValueTypes ) : Boolean; override;
    function SolveFunction( const FunctionName : string;
      Arguments : TStrings; var ReturnType : TValueType; var Return : string ) : Boolean; override;
  end;

  TForm1 = class(TForm)
    Edit1: TEdit;
    Memo1: TMemo;
    Memo3: TMemo;
    Memo2: TMemo;
    Label1: TLabel;
    Experssions: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    procedure Edit1KeyPress(Sender: TObject; var Key: Char);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    Calc : TTestCalculator;
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

function TTestCalculator.ResolveIdentifier( const Identifier : string;
  var ValueType : TValueType; var Value : string ) : Boolean;
begin
  Value := Trim( Form1.Memo2.Lines.Values[ Identifier ] );
  Result := True;
  if Length( Value ) = 0 then
    ValueType := vtEmpty
  else
    begin
      try
        StrToFloat( Value );
        ValueType := vtNumber;
      except
        ValueType := vtString;
      end;
    end;
end;

function TTestCalculator.LookupFunction( const FunctionName : string;
  var MinArgCount , MaxArgCount : Integer; var ValueTypes : TValueTypes ) : Boolean;
begin
  Result := inherited LookupFunction( FunctionName , MinArgCount , MaxArgCount , ValueTypes );
  if not Result then
    begin
      if EqualStr( FunctionName , 'AVERAGE' ) then
        begin
          MinArgCount := 1;
          MaxArgCount := High( Longint );
          ValueTypes := [ vtIdentifier , vtNumber , vtEmpty ];
          Result := True;
        end;
    end;
end;

function TTestCalculator.SolveFunction( const FunctionName : string;
  Arguments : TStrings; var ReturnType : TValueType; var Return : string ) : Boolean;
var
  Sum : Double;
  I : Integer;
begin
  Result := inherited SolveFunction( FunctionName , Arguments , ReturnType , Return );
  if not Result then
    begin
      if EqualStr( FunctionName , 'AVERAGE' ) then
        begin
          Result := True;
          ReturnType := vtNumber;
          ResolveArguments( Arguments , [ vtNumber , vtEmpty ] );
          ReplaceEmptysWith( Arguments , '0' );
          Sum := 0;
          for I := 0 to Pred( Arguments.Count ) do
            Sum := Sum + StrToFloat( Arguments[ I ] );
          Return := FloatToStr( Sum / Arguments.Count );
        end;
    end;
end;

procedure TForm1.Edit1KeyPress(Sender: TObject; var Key: Char);
var
  R : TValueType;
begin
  if Ord( Key ) = VK_RETURN then
    begin
      Key := #0;
      try
        Memo1.Lines.Insert( 0, Calc.Solve( Edit1.Text , R ) );
        Memo3.Lines.Insert( 0 , Edit1.Text );
      except on E : Exception do
        ShowMessage( E.Message );
      end;
    end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Calc := TTestCalculator.Create;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  Calc.Free;
end;

end.
